/*
 * Copyright (C) Sergey P. Derevyago, 2003-2004.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 *  ,   cexception.hpp.
 */

#include "cexception.hpp"
#include <exception>
#include <typeinfo>
#include "stringbuf.hpp"

using namespace std;

CException::CException(const FileLine& loc, const std::string& msg,
  sh_ptr<CException> nest) : location(loc), message(msg), nested(nest)
{
}

CException::~CException()
{
}

std::string CException::getClassName() const
{
 return "CException";
}

std::string CException::toString() const
{
 return StringBuf()+getClassName()+" ["+location.fname+":"+location.line+"]: "+
   message;
}

std::string CException::toStringAll() const
{
 string ret;
 ret.reserve(1024);

 for (const CException* ptr=this; ptr; ptr=ptr->nested.get()) {
     ret.append(ptr->toString());
     ret.push_back('\n');
 }

 return ret;
}

ExternalCException::ExternalCException(const FileLine& loc,
  const std::string& msg, const std::string& tname, sh_ptr<CException> nest) :
  CException(loc, msg, nest), typeName(tname)
{
}

ExternalCException::~ExternalCException()
{
}

std::string ExternalCException::getClassName() const
{
 return "ExternalCException";
}

std::string ExternalCException::toString() const
{
 return StringBuf()+getClassName()+" ["+location.fname+":"+location.line+
   "], typeName=\""+typeName+"\", message=\""+message+"\"";
}

STDExternalCException::STDExternalCException(const FileLine& loc,
  const std::string& msg, const std::string& tname, sh_ptr<CException> nest) :
  ExternalCException(loc, msg, tname, nest)
{
}

STDExternalCException::~STDExternalCException()
{
}

std::string STDExternalCException::getClassName() const
{
 return "STDExternalCException";
}

UnknownExternalCException::UnknownExternalCException(const FileLine& loc,
  const std::string& msg, const std::string& tname, sh_ptr<CException> nest) :
  ExternalCException(loc, msg, tname, nest)
{
}

UnknownExternalCException::~UnknownExternalCException()
{
}

std::string UnknownExternalCException::getClassName() const
{
 return "UnknownExternalCException";
}

sh_ptr<CException> toCException(const CException::FileLine& location)
{
#ifndef DERS_RETHROW_BUG  //  
 try { throw; }
 catch (sh_ptr<CException> ce) {
       return ce;
 }
 catch (const exception& e) {
       return newSTDExternalCException(location, e.what(), typeid(e).name());
 }
 catch (...) {
       return newUnknownExternalCException(location, "Unknown exception",
         "unknown");
 }
#else  //     
 if (CException::current.refs()>1 && CException::current.get())
    return CException::current;

 return newUnknownExternalCException(location, "Unknown exception", "unknown");
#endif
}

#ifdef DERS_RETHROW_BUG
sh_ptr<CException> CException::current;

sh_ptr<CException> newCException(const CException::FileLine& loc,
  const std::string& msg, sh_ptr<CException> nest)
{
 return CException::current=sh_ptr<CException>(new CException(loc, msg, nest));
}

sh_ptr<CException> newExternalCException(const CException::FileLine& loc,
  const std::string& msg, const std::string& tname, sh_ptr<CException> nest)
{
 return CException::current=sh_ptr<CException>(new ExternalCException(loc, msg,
   tname, nest));
}

sh_ptr<CException> newSTDExternalCException(
  const CException::FileLine& loc, const std::string& msg,
  const std::string& tname, sh_ptr<CException> nest)
{
 return CException::current=sh_ptr<CException>(new STDExternalCException(loc,
   msg, tname, nest));
}

sh_ptr<CException> newUnknownExternalCException(
  const CException::FileLine& loc, const std::string& msg,
  const std::string& tname, sh_ptr<CException> nest)
{
 return CException::current=sh_ptr<CException>(new UnknownExternalCException(
   loc, msg, tname, nest));
}
#endif
