/*
 * Copyright (C) Sergey P. Derevyago, 2003-2004.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 *   ,   .
 */

#ifndef __CEXCEPTION_HPP__
 #define __CEXCEPTION_HPP__

#include <string>
#include "config.hpp"
#include "sh_ptr.hpp"

/**
 *       .   
 *    sh_ptr<CException>,    
 *  ,   CException.  , 
 *         ,  
 *        " CException", 
 *        CException, : "
 *      ".
 */
class CException {
 public:
      /**
       *        .
       */
      struct FileLine {
             /**   */
             const char* fname;
             /**   */
             int line;

             /**
              *        .
              */
             FileLine(const char* f, int l) : fname(f), line(l) {}
      };

#ifdef DERS_RETHROW_BUG
      /**          */
      static sh_ptr<CException> current;
#endif

 private:
      /**   */
      CException(const CException&);
      /**   */
      CException& operator=(const CException&);

 protected:
      /**
       *  .     
       *  newCException().
       */
      CException(const FileLine& loc, const std::string& msg,
        sh_ptr<CException> nest);

 public:
      /**    */
      const FileLine location;
      /**  */
      const std::string message;
      /**    ,  ,  0 */
      const sh_ptr<CException> nested;

      /**
       *      .  
       *       ,  ,
       * ,  .    loc 
       *   _FLINE_.
       */
      friend sh_ptr<CException> newCException(const FileLine& loc,
        const std::string& msg, sh_ptr<CException> nest=sh_ptr<CException>());

      /**
       *  .
       */
      virtual ~CException();

      /**
       *   -.    
       *  .
       */
      virtual std::string getClassName() const;

      /**
       *  " " ,     
       * .    , 
       *  .
       */
      virtual std::string toString() const;

      /**
       *  " "     ,
       *   . ,   toString() 
       *       .
       */
      std::string toStringAll() const;

      /**
       *         
       *  E (   ).   
       *      .
       */
      template<class E>
      E* is() { return dynamic_cast<E*>(this); }

      /**
       *         
       *   E (   ).  
       *       .
       */
      template<class E>
      const E* is() const { return dynamic_cast<const E*>(this); }

      /**
       *      . 
       *     .
       */
      void* operator new(size_t size)
      {
       return sized_alloc::alloc(size);
      }

      /**
       *      . 
       *     .
       */
      void operator delete(void* ptr, size_t size)
      {
       sized_alloc::free(ptr, size);
      }
};

/**
 *      ,   
 *   ,  CException.
 */
class ExternalCException : public CException {
 protected:
      /**
       *  .     
       *  newExternalCException().
       */
      ExternalCException(const FileLine& loc, const std::string& msg,
        const std::string& tname, sh_ptr<CException> nest);
        
 public:
      /**
       *        .
       * ,     std::type_info::name().
       */
      const std::string typeName;

      /**
       *      .  
       *    , ,   , ,
       *  .    loc  
       *  _FLINE_.
       */
      friend sh_ptr<CException> newExternalCException(const FileLine& loc,
        const std::string& msg, const std::string& tname,
        sh_ptr<CException> nest=sh_ptr<CException>());

      /**
       *  .
       */
      virtual ~ExternalCException();

      /**
       *   -.    
       *  .
       */
      virtual std::string getClassName() const;

      /**
       *  " " ,     
       * .    , 
       *  .
       */
      virtual std::string toString() const;
};

/**
 *    std::exception  ,   .
 */
class STDExternalCException : public ExternalCException {
 protected:
      /**
       *  .     
       *  newSTDExternalCException().
       */
      STDExternalCException(const FileLine& loc, const std::string& msg,
        const std::string& tname, sh_ptr<CException> nest);

 public:
      /**
       *      .  
       *    , ,   , ,
       *  .    loc  
       *  _FLINE_.
       */
      friend sh_ptr<CException> newSTDExternalCException(const FileLine& loc,
        const std::string& msg, const std::string& tname,
        sh_ptr<CException> nest=sh_ptr<CException>());

      /**
       *  .
       */
      virtual ~STDExternalCException();

      /**
       *   -.    
       *  .
       */
      virtual std::string getClassName() const;
};

/**
 *      ,    
 * .
 */
class UnknownExternalCException : public ExternalCException {
 protected:
      /**
       *  .     
       *  newUnknownExternalCException().
       */
      UnknownExternalCException(const FileLine& loc, const std::string& msg,
        const std::string& tname, sh_ptr<CException> nest);

 public:
      /**
       *      .  
       *    , ,   , ,
       *  .    loc  
       *  _FLINE_.
       */
      friend sh_ptr<CException> newUnknownExternalCException(
        const FileLine& loc, const std::string& msg, const std::string& tname,
        sh_ptr<CException> nest=sh_ptr<CException>());

      /**
       *  .
       */
      virtual ~UnknownExternalCException();

      /**
       *   -.    
       *  .
       */
      virtual std::string getClassName() const;
};

/**
 *         
 *  sh_ptr<CException>.        catch. 
 *   location    _FLINE_.
 *      ()
 *        
 *   ,      
 *  ,   .
 */
sh_ptr<CException> toCException(const CException::FileLine& location);

/**
 *      CException::FileLine.
 */
#define _FLINE_ CException::FileLine(__FILE__, __LINE__)

#ifndef DERS_RETHROW_BUG
inline sh_ptr<CException> newCException(const CException::FileLine& loc,
  const std::string& msg, sh_ptr<CException> nest)
{
 return sh_ptr<CException>(new CException(loc, msg, nest));
}

inline sh_ptr<CException> newExternalCException(const CException::FileLine& loc,
  const std::string& msg, const std::string& tname, sh_ptr<CException> nest)
{
 return sh_ptr<CException>(new ExternalCException(loc, msg, tname, nest));
}

inline sh_ptr<CException> newSTDExternalCException(
  const CException::FileLine& loc, const std::string& msg,
  const std::string& tname, sh_ptr<CException> nest)
{
 return sh_ptr<CException>(new STDExternalCException(loc, msg, tname, nest));
}

inline sh_ptr<CException> newUnknownExternalCException(
  const CException::FileLine& loc, const std::string& msg,
  const std::string& tname, sh_ptr<CException> nest)
{
 return sh_ptr<CException>(new UnknownExternalCException(loc, msg, tname,
   nest));
}
#endif

#endif
