/*
 * Copyright (C) Sergey P. Derevyago, 2003-2004.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 *   FindFile.
 */

#include "findfile.hpp"
#include <dirent.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <list>
#include "stringbuf.hpp"
#include "namematcher.hpp"

using namespace std;

namespace {

 /**       */
 struct AutoDIR {
        /**    */
        DIR* d;

        /**      */
        AutoDIR(DIR* d_) : d(d_) {}
        /**   */
        ~AutoDIR() { closedir(d); }
 };

 /**
  * -  FindFile.     
  *     readdir(),   '/' 
  *     .
  */
 class FindFileImpl : public FindFile {
       /**
        *        
        *   NameMatcher.
        */
       virtual bool find(const std::string& dir, bool recur,
         const std::string& mask, FindFileCallback& cb);
 };

 bool FindFileImpl::find(const std::string& dir, bool recur,
   const std::string& mask, FindFileCallback& cb)
 {
  try {
      DIR *d=opendir(dir.c_str());
      if (!d) {
         throw newCException(_FLINE_, StringBuf("Can't open \"")+dir+
           "\" directory: "+strerror(errno));
      }
      AutoDIR autoDIR(d);

      typedef list<string> list_type;
      list_type subDirs;

      sh_ptr<NameMatcher> nm=Factory::newNameMatcher(mask);

      struct stat statbuf;
      dirent *ent;
      while (ent=readdir(d)) {
            if (!strcmp(ent->d_name,".") || !strcmp(ent->d_name,"..")) continue;

            string fname(StringBuf(dir)+"/"+ent->d_name);
            if (stat(fname.c_str(), &statbuf)) {
               throw newCException(_FLINE_, StringBuf("Can't stat \"")+fname+
                 "\": "+strerror(errno));
            }

            if (statbuf.st_mode&S_IFDIR) {
               subDirs.push_back(fname);
               continue;
            }

            if (!nm->match(ent->d_name)) continue;

            if (!cb.found(fname)) return 0;
      }

      if (recur) {
         for (list_type::const_iterator cit=subDirs.begin(), cend=subDirs.end();
              cit!=cend; ++cit) {
             if (!find(*cit, recur, mask, cb)) return 0;
         }
      }

      return 1;
  }
  catch (...) {
        throw newCException(_FLINE_, StringBuf("Can't process \"")+dir+
          "\" directory", toCException(_FLINE_));
  }
 }

}

sh_ptr<FindFile> Factory::newFindFile()
{
 try { return sh_ptr<FindFile>(new FindFileImpl); }
 catch (...) {
       throw newCException(_FLINE_, "Can't create FindFileImpl object",
         toCException(_FLINE_));
 }
}
