/*
 * Copyright (C) Sergey P. Derevyago, 2003-2004.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 *    ftext.   main()    
 *  .
 */

#include <stdio.h>
#include "findtext.hpp"
#include "stringbuf.hpp"

using namespace std;

/**
 *         .
 */
class Finder : public Subscriber<FoundMsg>, public Subscriber<ErrorMsg> {
      /**
       *   .
       */
      virtual bool regularMsg(const Publisher<FoundMsg>& pub,
        const FoundMsg& msg);

      /**
       *    .
       */
      virtual bool regularMsg(const Publisher<ErrorMsg>& pub,
        const ErrorMsg& msg);
};

bool Finder::regularMsg(const Publisher<FoundMsg>&, const FoundMsg& msg)
{
 printf("%s:%d:%s\n", msg.fileName.c_str(), msg.lineNum, msg.lineText.c_str());
 return 1;
}

bool Finder::regularMsg(const Publisher<ErrorMsg>&, const ErrorMsg& msg)
{
 fprintf(stderr, "%s\n", msg.errMsg.c_str());
 return 1;
}

/**
 *           
 * .
 */
class Logger : public Subscriber<FoundMsg>, public Subscriber<ErrorMsg>,
  public Subscriber<InfoMsg> {
      /**    */
      FILE* log;

      /**
       *   .
       */
      virtual bool regularMsg(const Publisher<FoundMsg>& pub,
        const FoundMsg& msg);

      /**
       *    .
       */
      virtual bool regularMsg(const Publisher<ErrorMsg>& pub,
        const ErrorMsg& msg);

      /**
       *   .
       */
      virtual bool regularMsg(const Publisher<InfoMsg>& pub,
        const InfoMsg& msg);

 public:
      /**  ,     */
      int filesOpened;
      /**  ,     */
      int filesClosed;
      /**    */
      int linesFound;
      /**    */
      int errorsEncountered;

      /**
       *        .
       */
      Logger(const string& fname);

      /**
       *        .
       */
      ~Logger();
};

Logger::Logger(const string& fname)
{
 log=fopen(fname.c_str(), "a");
 if (!log) {
    throw newCException(_FLINE_, StringBuf("Can't open \"")+fname+
      "\" file for logging");
 }

 filesOpened=filesClosed=linesFound=errorsEncountered=0;
}

Logger::~Logger()
{
 fputc('\n', log);
#define PR(var) fprintf(log, #var ": %d\n", var)
 PR(filesOpened);
 PR(filesClosed);
 PR(linesFound);
 PR(errorsEncountered);
#undef PR
 fputc('\n', log);

 fclose(log);
}

bool Logger::regularMsg(const Publisher<FoundMsg>& pub, const FoundMsg& msg)
{
 linesFound++;
 fprintf(log, "FoundMsg: Publisher(%p) FindText(%p) fileName(%s) lineNum(%d) "
   "lineText(%s)\n", &pub, &msg.ft, msg.fileName.c_str(), msg.lineNum,
   msg.lineText.c_str());

 return 1;
}

bool Logger::regularMsg(const Publisher<ErrorMsg>& pub, const ErrorMsg& msg)
{
 errorsEncountered++;
 fprintf(log, "ErrorMsg: Publisher(%p) FindText(%p) errMsg(%s)\n", &pub,
   &msg.ft, msg.errMsg.c_str());

 return 1;
}

bool Logger::regularMsg(const Publisher<InfoMsg>& pub, const InfoMsg& msg)
{
 if (msg.type==InfoMsg::opened) filesOpened++;
 else filesClosed++;

 fprintf(log, "InfoMsg : Publisher(%p) FindText(%p) type(%s) fileName(%s)\n",
   &pub, &msg.ft, (msg.type==InfoMsg::opened) ? "opened" : "closed",
   msg.fileName.c_str());

 return 1;
}

/**
 *       .
 */
struct CmdLineParser {
       /**  -l */
       bool isL;
       /**  -,   isL */
       string logName;
       /**  -r */
       bool isR;
       /**      */
       string dirName;
       /**     */
       string mask;
       /**      */
       string text;

       /**
        *      .
        */
       CmdLineParser() : isL(0), isR(0) {}

       /**
        *    
        * "[-l logname] [-r] dirname mask text"
        *    .     
        * false.
        */
       bool parse(int argc, char** argv);
};

bool CmdLineParser::parse(int argc, char** argv)
{
 if (argc<4 || argc>7) return 0;

 int curr=1;
 while (argv[curr][0]=='-') {
       string opt(argv[curr]);

       if (opt=="-l") {
          if (isL) return 0;

          isL=1;
          curr++;

          logName=argv[curr++];
       }
       else if (opt=="-r") {
          if (isR) return 0;

          isR=1;
          curr++;
       }
       else return 0;
 }

 if (curr+3!=argc) return 0;

 dirName=argv[curr++];
 mask=argv[curr++];
 text=argv[curr++];

 return 1;
}

/**
 *      .
 */
void printUsage()
{
 fprintf(stderr,
"Usage:\n"
"  ftext [-l logname] [-r] dirname mask text\n"
"For example:\n"
"  ftext -l ftext.log -r . *pp return\n"
 );
}

int main(int argc, char** argv)
{
 try {
     CmdLineParser clp;
     if (!clp.parse(argc, argv)) {
        //  /  "*pp"    ,
        //   "pp".     
        //     .
        //   /  "*pp"   
        // "pp", ..   "*".
        printf("Invalid command line:\n");
        for (int i=0; i<argc; i++)
            printf("%s%s", argv[i], (i!=argc-1)? " " : "\n\n");

        printUsage();
        return 1;
     }

     sh_ptr<FindText> ft=Factory::newFindText();

     Finder f;
     ft->foundPub().subscribe(f);
     ft->errorPub().subscribe(f);

     sh_ptr<Logger> l;
     if (clp.isL) {
        l.set(new Logger(clp.logName));

        ft->foundPub().subscribe(*l);
        ft->errorPub().subscribe(*l);
        ft->infoPub().subscribe(*l);
     }
     
     ft->search(clp.dirName, clp.isR, clp.mask, clp.text);

     return 0;
 }
 catch (...) {
       fprintf(stderr, "Uncaught exception:\n%s",
         toCException(_FLINE_)->toStringAll().c_str());
       return 2;
 }
}
